from unittest import TestCase

from appyx.layers.domain.operations.domain_operation import DomainOperation


class DomainOperationTest(TestCase):
    """

    """

    def test_operation_has_error_when_the_string_doesnt_exist_in_collection(self):
        # Given a new operation
        operation = PrefixStringWithTest(["test_hereje", "qwerty"], "sarasa")

        # When we execute it
        result = operation.execute()

        # Then we get an abort return result
        self.assertTrue(result.has_errors())
        operation_error = result.errors()[0]
        self.assertEqual('El string no está', operation_error.text())

    def test_operation_ends_successfully_when_the_string_is_in_collection_with_prefix_already_attached(self):
        # Given a new operation
        operation = PrefixStringWithTest(["test_hereje", "qwerty"], "hereje")

        # When we execute it
        result = operation.execute()

        # Then we get an early return result
        self.assertTrue(result.is_successful())
        self.assertEqual("test_hereje", result.get_object())

    def test_attaches_prefix_successfully_when_the_string_is_in_collection(self):
        # Given a new operation
        operation = PrefixStringWithTest(["test_hereje", "qwerty"], "qwerty")

        # When we execute it
        result = operation.execute()

        # Then we get a successful result
        self.assertTrue(result.is_successful())
        self.assertEqual("test_qwerty", result.get_object())


class PrefixStringWithTest(DomainOperation):
    def __init__(self, a_string_collection, a_string_to_check):
        super().__init__()
        self._prefix = 'test_'
        self._string_collection = a_string_collection
        self._string_to_check = a_string_to_check

    def _execute_logic(self):

        matched_string = None
        for a_string in self._string_collection:
            if a_string.endswith(self._string_to_check):
                matched_string = a_string

        if matched_string is None:
            self._result.add_error('El string no está')
        else:
            self._result.set_object(self._add_test_prefix(matched_string))

    def _add_test_prefix(self, a_string):
        if a_string.startswith(self._prefix):
            return a_string
        return f'{self._prefix}{a_string}'



import datetime
from decimal import Decimal
from typing import Iterable

from appyx.layers.application.interactions.validators.parameters import ParameterValidator
from appyx.layers.domain.result import Result


class NotNone(ParameterValidator):

    def validate(self, argument, result=None, parameter_name=None):
        """Nota: Sobrecargamos el método padre porque no queremos que nos puenteen la validacion si el argumento es None
        """
        if result is None:
            result = Result()
        return self.validate_with_result(argument, result, parameter_name)

    def _validation_result(self, argument):
        validation_result = Result()
        condition = argument is not None
        if not condition:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo no puede estar vacío'


class NotEmptyString(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        condition = argument != ""
        if not condition:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo no puede estar en blanco'


class IsString(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        condition = isinstance(argument, str)
        if not condition:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser un string'


class IsAPositiveInteger(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        condition = isinstance(argument, int) and argument >= 0

        if not condition:
            validation_result.add_error(self.error_message())

        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser un entero positivo'


class IsBoolean(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        condition = isinstance(argument, bool)
        if not condition:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser un booleano'


class IsList(ParameterValidator):
    def __init__(self, item_validators=None) -> None:
        super().__init__()
        self._item_validators = item_validators or []

    def _validation_result(self, argument: Iterable):
        validation_result = Result()
        condition = isinstance(argument, list)
        if not condition:
            validation_result.add_error(self.error_message())
            return validation_result

        for list_item in argument:
            for validator in self._item_validators:
                validator.validate(list_item, validation_result)

        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser una lista'


class IsDictionary(ParameterValidator):
    """
    Note: we could add support to validate keys and/or values
    """

    def __init__(self) -> None:
        super().__init__()

    def _validation_result(self, argument: Iterable):
        validation_result = Result()
        condition = isinstance(argument, dict)
        if not condition:
            validation_result.add_error(self.error_message())
            return validation_result

        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser un diccionario'


class IsAnID(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        is_int_gte_to_1 = isinstance(argument, int) and argument >= 1
        is_str_gte_to_1 = isinstance(argument, str) and argument.isdigit() and int(argument) >= 1
        if not (is_int_gte_to_1 or is_str_gte_to_1):
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser un ID válido'


class IsAFiniteDecimal(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        condition = isinstance(argument, Decimal) and argument.is_finite()
        if not condition:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo debe ser un número válido'


class NotEmpty(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        if len(argument) == 0:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'Este campo no puede ser vacio'


class AllIncluded(ParameterValidator):

    def __init__(self, items) -> None:
        super().__init__()
        self._items = items

    @classmethod
    def new_with(cls, items):
        return cls(items)

    def _validation_result(self, argument):
        validation_result = Result()
        not_included = [each for each in argument if each not in self._items]
        if not_included:
            not_included_as_string = ", ".join(not_included)
            validation_result.add_error(f"{not_included_as_string} are not valid items")
        return validation_result


class IsADateString(ParameterValidator):
    def __init__(self, pattern='%Y-%m-%dT%H:%M:%S.%fZ'):
        self._pattern = pattern

    def _validation_result(self, argument):
        validation_result = Result()
        try:
            datetime.datetime.strptime(argument, self._pattern)
        except (TypeError, ValueError):
            validation_result.add_error(self.error_message(self._pattern))
        return validation_result

    @classmethod
    def error_message(cls, pattern):
        return f'Este campo debe ser una fecha válida con formato "{pattern}"'


class IsTelephoneNumber(ParameterValidator):
    def _validation_result(self, argument):
        validation_result = Result()
        if type(argument) is not str:
            validation_result.add_error("Field is not a string")
            return validation_result

        condition = len(argument) >= 8
        if not condition:
            validation_result.add_error(self.error_message())
        return validation_result

    @classmethod
    def error_message(cls):
        return 'El teléfono es inválido'


class IsEmailAddress(ParameterValidator):
    """Modela la validacion de si una string es una direccion valida de email.

    Nota: acoplados con Django
    Nota2: Existe un package "validate_email" que hace precisamente esto.
    """

    def _validation_result(self, argument):
        validation_result = Result()

        from django.core.validators import EmailValidator
        from django.core.exceptions import ValidationError
        try:
            EmailValidator().__call__(argument)
        except ValidationError as exception:
            validation_result.add_error(str(exception.messages))
        return validation_result


class IsValidContentType(ParameterValidator):
    def __init__(self, content_types):
        self._valid_content_types = content_types

    def _validation_result(self, argument):
        validation_result = Result()
        content_type = argument.content_type
        if content_type not in self._valid_content_types:
            valid_content_types = ", ".join(str(content_type) for content_type in self._valid_content_types)
            validation_result.add_error(
                f'The content type "{content_type}" is not supported. It should be one of "{valid_content_types}"')

        return validation_result

    @classmethod
    def with_content_types(cls, content_types):
        return cls(content_types)


class MinimumValue(ParameterValidator):
    def __init__(self, value):
        self._value = value

    def _validation_result(self, argument):
        validation_result = Result()
        if argument < self._value:
            validation_result.add_error(f'Value must be greater or equal than {self._value}. Value was {argument}.')

        return validation_result


class MaximumValue(ParameterValidator):
    def __init__(self, value):
        self._value = value

    def _validation_result(self, argument):
        validation_result = Result()
        if argument > self._value:
            validation_result.add_error(f'Value must be less or equal than {self._value}. Value was {argument}.')

        return validation_result


class MinimumLength(ParameterValidator):
    def __init__(self, min_length):
        self._min_length = min_length

    def _validation_result(self, argument):
        validation_result = Result()
        argument_length = len(argument)
        if argument_length < self._min_length:
            validation_result.add_error(
                f'Value must be of longer or equal length than {self._min_length} characters.'
                f'Value was {argument_length} characters long.')

        return validation_result


class IsOneOfThese(ParameterValidator):
    def __init__(self, list_of_valid_elements):
        self._list_of_valid_elements = list_of_valid_elements

    def _validation_result(self, element):
        validation_result = Result()
        is_one_of_these = element in self._list_of_valid_elements
        if not is_one_of_these:
            valid_elements = ", ".join(str(element) for element in self._list_of_valid_elements)
            validation_result.add_error(f'Value "{element}" must be one of these elements: "{valid_elements}"')

        return validation_result


class MaximumLength(ParameterValidator):
    def __init__(self, max_length):
        self._max_length = max_length

    def _validation_result(self, argument):
        validation_result = Result()
        argument_length = len(argument)
        if self._max_length < argument_length:
            validation_result.add_error(
                f'Value must be shorter or equal length than {self._max_length} characters.'
                f'Value was {argument_length} characters long.')

        return validation_result



from appyx.layers.domain.business import Business


class PollStation(Business):
    def __init__(self, clock, questions_archive):
        self._clock = clock
        self._questions = questions_archive

    def clock(self):
        return self._clock

    def latest_questions(self):
        return self._questions.latest(5)

    def create_question_with_choices(self, question_text, choices_texts):
        new_question = Question(question_text, choices_texts, self._clock.current_time())
        self._questions.add(new_question)

    def vote(self, question, choice):
        if choice not in question.choices():
            raise AssertionError('Choice must belong to question')
        choice.increment_votes()


class Choice:
    def __init__(self, text, votes):
        self._text = text
        self._votes = votes

    def text(self):
        return self._text

    def amount_of_votes(self):
        return self._votes

    def increment_votes(self):
        self._votes = self._votes + 1


class Question:
    def __init__(self, text, choices_texts, creation_time):
        self._text = text
        self._choices = [Choice(text, 0) for text in choices_texts]
        self._creation_time = creation_time

    def text(self):
        return self._text

    def choices(self):
        return self._choices

    def choices_texts(self):
        return [choice.text() for choice in self.choices()]

    def creation_time(self):
        return self._creation_time


class FakeClock:
    def __init__(self):
        self._current_time = 1

    def current_time(self):
        return self._current_time

    def advance_1_day(self):
        self._current_time = self._current_time + 1


class SystemClock:
    def __init__(self):
        import datetime
        self._clock = datetime.datetime

    def current_time(self):
        return self._clock.now()

    def advance_1_day(self):
        raise NotImplementedError('Should not implement')


class QuestionsArchive:
    def __init__(self):
        self._objects = []

    def latest(self, an_amount):
        answer_list = self._objects.copy()
        answer_list.sort(key=lambda question: question.creation_time())
        return answer_list[:an_amount]

    def add(self, question):
        self._objects.append(question)



from unittest import TestCase

from appyx.layers.application.interactions.app_interaction import AppInteraction
from appyx.layers.application.interactions.parameters.test_command_parameters import TestApp


class AppInteractionWhiteBoxTest(TestCase):
    """
    Tests the implementation of the abstract class AppInteraction (ex AppCommand)
    These tests are meant to be run on appyx alone.
    It guarantees that after a refactor in appyx is made, appyx itself can check it's still working properly.
    Subclasses of AppInteraction are not required to run these tests
    since they can override the implementation of the private methods tested here.
    """

    # From here we test the implementation of the template method execute defined in AppInteraction

    def test_the_result_of_execute_is_the_result_from_execute_from_successful_result(self):
        """
            When an interaction has no errors during the validate()
            the result of execute is the result from _execute_from_successful_result()
        """
        # Given
        interaction = self._new_sample_app_interaction()

        # When
        result = interaction.execute()

        # Then
        self.assertTrue(result.is_successful())
        self.assertEqual(result.get_object(), self._hello_world())

    def _new_sample_app_interaction(self):
        hello_world = self._hello_world()

        class HelloWorldInteraction(AppInteraction):
            def _execute_from_successful_parameter_validation(self, result):
                result = super()._execute_from_successful_parameter_validation(result)
                result.set_object(hello_world)
                return result

        return HelloWorldInteraction(TestApp())

    def _hello_world(self):
        return "Hello World"



class HTTPInterface:
    def __init__(self, name, base_url, default_authentication_methods):
        self._name = name
        self._base_url = base_url
        self._default_authentication_methods = default_authentication_methods
        self._sections = self._initialize_sections()

    def name(self):
        return self._name

    def url(self):
        return f'{self._base_url}/{self._route_prefix()}'

    def description(self):
        raise NotImplementedError("Subclass responsibility")

    def version(self):
        raise NotImplementedError("Subclass responsibility")

    def sections(self):
        return self._sections

    def endpoints(self):
        return self._concatenate_all_endpoints_for_all_sections()

    def default_authentications_methods(self):
        return self._default_authentication_methods

    def authentication_methods_for_endpoint_named(self, endpoint_name):
        security_methods = self._authentication_methods_for_excepted_endpoint_named(endpoint_name)
        if security_methods is None:
            security_methods = self.default_authentications_methods()
        return security_methods

    def route_for_endpoint_named(self, endpoint_name, kwargs=None):
        if kwargs is None:
            kwargs = {}

        return f"/{self._route_prefix()}/{self.endpoint_named(endpoint_name).route_with_variable_replacements(kwargs)}"

    def route_pattern_for_endpoint_named(self, endpoint_name):
        return f"{self._route_prefix()}/{self.relative_route_pattern_for_endpoint_named(endpoint_name)}"

    def relative_route_pattern_for_endpoint_named(self, endpoint_name):
        return f"{self.endpoint_named(endpoint_name).route()}"

    def endpoint_named(self, endpoint_name):
        endpoints_with_name = [endpoint for endpoint in self.endpoints() if endpoint.name() == endpoint_name]
        self._validate_endpoint_exists_and_is_unique(endpoints_with_name, endpoint_name)
        return endpoints_with_name[0]

    def _initialize_sections(self):
        raise NotImplementedError("Subclass responsibility")

    def _authentication_methods_for_excepted_endpoint_named(self, endpoint_name):
        return self._authentication_methods_exceptions().get(endpoint_name)

    def _authentication_methods_exceptions(self):
        return {}

    def _concatenate_all_endpoints_for_all_sections(self):
        return [endpoint for section in self.sections() for endpoint in section.endpoints()]

    def _validate_endpoint_exists_and_is_unique(self, endpoint_list, endpoint_name):
        if len(endpoint_list) == 0:
            raise Exception(f"Endpoint named {endpoint_name} not found in interface named {self.name()}")
        elif len(endpoint_list) > 1:
            raise Exception(f"Multiple endpoints named {endpoint_name} found in interface named {self.name()}")

    def _route_prefix(self):
        raise NotImplementedError("Subclass responsibility")



from appyx.layers.domain.exceptions.early_return import EarlyReturnBecauseOfErrorsException
from appyx.layers.domain.result import Result


class DomainOperation:
    """
    Models an operation from the Domain.
    Answers a Result after executing the operation.
    Can delegate the return from execute() to private methods.

    Subclasses should implement _execute_logic(), calling _return() if they need to return from execute().

    Note: _save_model is a crutch that couples us to Django. Further iterations will remove this coupling.
    """
    def __init__(self) -> None:
        super().__init__()
        self._result = Result()

    def execute(self):
        self._result = Result()

        try:
            self._execute_logic()
        except EarlyReturnBecauseOfErrorsException:
            pass

        return self._result

    def _abort_in_case_of_errors(self):
        # If the result has errors, exit the operation.
        # Otherwise continue executing normally
        if self._result.has_errors():
            self._return_from_operation()

    def _return_from_operation(self):
        raise EarlyReturnBecauseOfErrorsException()

    def _execute_logic(self):
        raise NotImplementedError("Subclass responsibility")



from appyx.layers.application.interactions.interaction import Interaction
from appyx.layers.application.interactions.parameters.interaction_parameters import InteractionParameters
from appyx.layers.domain.exceptions.early_return import EarlyReturnBecauseOfErrorsException
from appyx.layers.domain.result import Result


class AppInteraction(Interaction):
    """
    I model an interaction in the context of an Application.
    I usually model also the operation that is being performed.
    """

    def __init__(self, app):
        super().__init__()
        self._app = app
        self._parameters = InteractionParameters(self,
                                                 self._initial_parameters(),
                                                 self._initial_required_parameter_names(),
                                                 self._initial_at_least_one_required_sets(),
                                                 self._default_arguments())

    def app(self):
        return self._app

    def set_arguments(self, arguments):
        """
        Set the values that each parameter will have.
        Arguments is a dictionary with the names of the parameters as keys and the value of the argument as value.

        :param arguments: dict of str: object
        """
        self._parameters.set_arguments(arguments)

    def set_keyword_arguments(self, **arguments):
        """
        Set the values that each parameter will have.
        Arguments are keyword arguments.
        """
        # TODO: @laski thinks this should be the only way to set arguments.
        # Compare:
        #     command.set_arguments({command.PARAMETER_KEY_USER: buenbit_user})
        # with
        #     command.set_arguments(user=buenbit_user)
        # The latter is clearer, easier to remember and delegates to Python the validation of the arguments.
        # Also, using string as argument values is error
        self.set_arguments(arguments)

    def can_execute(self):
        """
        Answers the result of trying to execute.

        If result has no errors, it means the command can be executed (although it may fail if further problems arise).
        If result has errors, then the command can not be executed (and the errors explain why)
        """
        result = Result()
        try:
            result = self._validate_parameters_and_arguments(result)
            return result
        except Exception as exc:
            exception_handler = self.app().general_exception_handler()
            result = exception_handler.handle(exc, result)
            return result

    def validate(self):
        return self.can_execute()

    def execute(self):
        """
        Performs the operation of the receiver and answers the result of the execution.

        :return: result of execution
        """
        result = self.can_execute()
        try:
            if result.has_errors():
                return result
            return self._execute_from_successful_parameter_validation(result)
        except EarlyReturnBecauseOfErrorsException:
            return result
        except Exception as exc:
            exception_handler = self.app().general_exception_handler()
            result = exception_handler.handle(exc, result)
            return result

    def execute_with(self, **kwargs):
        """
        Syntax sugar to simplify command execution with a list of arguments
        Replaces boilerplate constructions like this:
            arguments = {ValidateUserSafetyCommand.PARAMETER_KEY_USER: buenbit_user}
            command.set_arguments(arguments)
            result = command.execute()
        with:
            result = command.execute_with(user=buenbit_user)
        """
        self.set_arguments(kwargs)
        result = self.execute()
        return result

    def get_parameters(self):
        return self._parameters

    def add_parameter(self, parameter):
        self._parameters.add_parameter(parameter)

    def set_parameters(self, parameters):
        self._parameters.clear_parameters()
        self.add_parameters(parameters)

    def add_parameters(self, parameters):
        for parameter in parameters:
            self.add_parameter(parameter)

    def set_argument_value(self, argument_name, value):
        return self.get_parameters().set_argument_value(argument_name, value)

    def is_defined_argument_named(self, argument_name):
        return self.get_parameters().is_defined_argument_named(argument_name)

    def get_argument_named(self, argument_name):
        return self.get_parameters().get_argument_named(argument_name)

    def get_parameter_named(self, name):
        return self._parameters.get_parameter_named(name)

    def add_parameters_validator(self, parameters_validator):
        self._parameters.add_parameters_validator(parameters_validator)

    def add_required_parameter_named(self, parameter_name):
        self._parameters.add_required_parameter_named(parameter_name)

    def remove_required_parameter_named(self, parameter_name):
        self._parameters.remove_required_parameter_named(parameter_name)

    def required_parameter_names(self):
        return self._parameters.required_parameter_names()

    def add_at_least_one_constraint(self, parameter_names):
        self._parameters.add_at_least_one_constraint(parameter_names)

    def add_at_least_one_not_none_constraint(self, parameter_names):
        self._parameters.add_at_least_one_not_none_constraint(parameter_names)

    def handle_unexpected_arguments(self, extra_arguments, result):
        """Handle the existence of more arguments than those supported by parameters.

        Note: Currently, we don't raise warnings or errors for this case.
        If you want to do something about extra arguments, override this method.

        :param extra_arguments: the names of the arguments that has no matching parameter
        :param result: result object in which to add the error or warning about this issue
        """
        # TODO: Make default handling to add warnings to the result for these extra_arguments
        pass

    # --- private methods ---

    def _abort_in_case_of_errors(self, result):
        if result.has_errors():
            self._return_from_command()

    def _return_from_command(self):
        raise EarlyReturnBecauseOfErrorsException()

    def _initial_required_parameter_names(self):
        """
        Answer the names of the required parameters.
        If my subclass has one or more required parameters,
        it should override me answering the names of that/those parameter/s.

        :return: set of str
        """
        return set()

    def _execute_from_successful_parameter_validation(self, result):
        """
        My subclass should override me. Performs the operation and answers the result of the execution

        :return: result of execution
        """
        return result

    def _initial_parameters(self):
        """
        If I have more than 0 parameters, I should answer the definition of that/those parameter/s.

        :rtype: list of Parameter
        """
        return []

    def _initial_at_least_one_required_sets(self):
        """Answer a list of sets. Each set with the names of parameters from which al least one is required"""
        return []

    def _default_arguments(self):
        """Answer a dictionary with the default command arguments"""
        return {}

    def _validate_parameters_and_arguments(self, result):
        return self._parameters.validate_parameters_and_arguments(result)



class BaseError:
    """
    Models a generic, basic error.
    Provides information on the type of error (code) and details about it (object).

    code: indicates the class or type of error.
    object: provides the details of the error in a way determined by the code.
    """
    def __init__(self, data) -> None:
        super().__init__()
        self._object = data

    def object(self):
        return self._object

    def code(self):
        return self.__class__.__name__

    def __str__(self):
        return f'{self.__class__.__name__}<Data "{str(self.object())}">'

    def __repr__(self):
        return f'{self.__class__.__name__}(_object="{str(self.object())}")'

    def __unicode__(self):
        return self.__str__()


class SimpleError(BaseError):
    """
    Models a simple error.
    Provides the type of error (code) and a description of it (text).

    code: indicates the class or type of error.
    object: provides a description of the error.
    """
    def __init__(self, text, code) -> None:
        super().__init__(text)
        self._code = code

    def code(self):
        return self._code

    def text(self):
        return self.object()

    def __str__(self):
        return f'{self.__class__.__name__}<Code "{self.code()}" | Data "{str(self.object())}">'

    def __repr__(self):
        return f'{self.__class__.__name__}(_code="{self.code()}", _object="{str(self.text())}")'

    def __unicode__(self):
        return self.__str__()

    def __eq__(self, other):
        return super.__eq__(self, other) and self.code() == other.code() and self.text() == other.text()

    def __hash__(self):
        return hash((self._code, self._object))


class UnexpectedError(SimpleError):
    """
    Models an error that has been detected and caught
    but one which the system is not prepared to handle.

    The text of the error is expected to describe what happened and why it wasn't expected.
    """
    pass


class ParameterValidationError(SimpleError):
    """
    Models the invalidity of an argument in the context of a parameter.
    It provides the name of the parameter, the value of the argument
    and the different errors/reasons of why it is invalid.
    """
    def __init__(self, parameter_name, argument, validation_errors, code) -> None:
        self._parameter_name = parameter_name
        self._argument = argument
        self._validation_errors = validation_errors
        super().__init__(self._error_text(), code)

    def parameter_name(self):
        return self._parameter_name

    def argument(self):
        return self._argument

    def validation_errors(self):
        return self._validation_errors

    def validation_errors_text(self):
        return " | ".join(validation_error.text() for validation_error in self._validation_errors)

    def _error_text(self):
        return f'Parameter "{self._parameter_name}" is not valid with argument "{self._argument}". ' \
               f'Reasons: {self.validation_errors_text()}'


class ParametersValidationError(SimpleError):
    """
    Models the invalidity of a set of arguments in the context of a set of parameters.
    It provides the set of arguments, parameters and errors/reasons of the invalidity.
    """

    def __init__(self, parameters_and_arguments, validation_errors, code) -> None:
        self._parameters_and_arguments = parameters_and_arguments
        self._validation_errors = validation_errors
        super().__init__(self._error_text(), code)

    def parameters_and_arguments(self):
        return self._parameters_and_arguments

    def validation_errors(self):
        return self._validation_errors

    def validation_errors_text(self):
        return " | ".join(validation_error.text() for validation_error in self._validation_errors)

    def _error_text(self):
        keys = ", ".join(str(key) for key in self._parameters_and_arguments.keys())
        values = ", ".join(str(value) for value in self._parameters_and_arguments.values())
        return f'Parameters "{keys}" are not valid with arguments "{values}". Reasons: {self.validation_errors_text()}'


class BaseErrorFactory:
    """
    Models a Factory (design pattern) of commonly used errors.
    """
    def simple_error(self, text_or_exception):
        return self._new_error(text_or_exception, SimpleError, self.simple_error_code())

    def unexpected_error(self, text_or_exception):
        return self._new_error(text_or_exception, UnexpectedError, self.unexpected_error_code())

    def parameter_validation_error(self, parameter_name, argument, validation_errors):
        return ParameterValidationError(parameter_name, argument, validation_errors,
                                        self.parameter_validation_error_code())

    def parameters_validation_error(self, parameters_and_arguments, validation_errors):
        return ParametersValidationError(parameters_and_arguments, validation_errors,
                                         self.parameters_validation_error_code())

    def parameter_missing_error(self, text_or_exception):
        return self._new_error(text_or_exception, SimpleError, self.parameter_missing_error_code())

    def _new_error(self, text_or_exception, error_class, error_code):
        error_message = str(text_or_exception)
        return error_class(error_message, error_code)

    def simple_error_code(self):
        return "simple_error_code"

    def unexpected_error_code(self):
        return "unexpected_error_code"

    def parameter_validation_error_code(self):
        return "parameter_validation_error_code"

    def parameter_missing_error_code(self):
        return "parameter_missing_error"

    def parameters_validation_error_code(self):
        return "parameters_validation_error_code"



from typing import Dict

from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from appyx import Result
from appyx.layers.interface.exceptions.parameters_required import ParametersRequiredException
from appyx.layers.interface.handler_parameters import HandlerParameters
from appyx.layers.interface.presenters.standard_response_content import StandardResponseContentPresenter


class AppyxHttpHandler(APIView):
    """
    Models a HTTP request handler for an application made by Eryx.
    Defines abstract methods on how to handle it.

    Assumes that a interaction will be responsible for performing the logic of the request.
    Handles a basic parameter check prior to delegating the proper handling of parameters to the interaction.
    """

    # class attributes is the way django allows us to pass parameters to view instances through `as_view()` and
    # receive them as kwargs on the __init__
    authenticators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._authenticators = kwargs.get("authenticators", [])
        self._parameters = HandlerParameters(initial_parameters=self._initial_parameters(),
                                             initial_required_parameter_names=self._initial_required_parameter_names())

    # --- handling ---

    def handle_get(self, request, **kwargs):
        interaction = self._get_interaction_for_GET(request, kwargs)
        result = interaction.execute()
        return self._response_for(result)

    def handle_post(self, request, **kwargs):
        try:
            self._check_for_required_parameters_in(request)
            interaction = self._get_interaction_for_POST(request, kwargs)
            result = interaction.execute()
            response = self._response_for(result)
            self._notify_post_request(request, interaction, result)
            return response
        except ParametersRequiredException as exception:
            error_from_exception = self._error_factory().request_error(error_text=str(exception))
            result = self._result_with_error(error_from_exception)
            return Response(status=status.HTTP_400_BAD_REQUEST, data=self._trimmed_result(result))

    # --- authentication ---

    def get_authenticators(self):
        return self._authenticators

    # --- self description ---

    def description(self):
        return "Description not implemented. Override method `description()` in your handler class"

    def parameters(self):
        """
        :return: the parameter list
        """
        return self._parameters.parameters()

    def required_parameter_names(self):
        return self._parameters.required_parameter_names()

    # --- private methods ---

    def _error_factory(self):
        raise NotImplementedError('Subclass responsibility')

    def _get_interaction_for_GET(self, request, kwargs):
        raise NotImplementedError('Subclass responsibility')

    def _check_for_required_parameters_in(self, request):
        parameters = self._parameters_from(request)
        missing_parameters = []
        for required_parameter in self._parameters.required_parameter_names():
            if required_parameter not in parameters.keys():
                missing_parameters.append(required_parameter)

        if missing_parameters:
            raise ParametersRequiredException('%s parameters are required.' % missing_parameters)

    def _initial_required_parameter_names(self):
        return set()

    def _initial_parameters(self):
        return []

    def _parameters_from(self, request):
        return request.data or {}

    def _status_http_code_for(self, result):
        if result.is_successful():
            return self._status_code_for_successful_result(result)
        else:
            return self._status_code_for_unsuccessful_result(result)

    def _get_interaction_for_POST(self, request, kwargs):
        raise NotImplementedError('Subclass responsibility')

    def _status_code_for_unsuccessful_result(self, result):
        raise NotImplementedError('Subclass responsibility')

    def _status_code_for_successful_result(self, result):
        raise NotImplementedError('Subclass responsibility')

    def _response_for(self, result):
        status_http_code = self._status_http_code_for(result)
        data = self._response_content_for(result)
        return Response(status=status_http_code, data=data)

    def _trimmed_errors(self, result):
        trimmed_errors = []
        errors = result.errors()
        error_renderer = self._error_renderer()
        for error in errors:
            trimmed_error = error_renderer.render(error)
            trimmed_errors.append(trimmed_error)

        return trimmed_errors

    def _trimmed_result(self, result):
        shallow_object = self._trimmed_object(result)
        shallow_errors = self._trimmed_errors(result)
        return StandardResponseContentPresenter(an_object=shallow_object, errors=shallow_errors).present()

    def _trimmed_object(self, result):
        raise NotImplementedError('Subclass responsibility')

    def _error_renderer(self):
        raise NotImplementedError('Subclass responsibility')

    def _response_content_for(self, result):
        response_content = self._default_response_content()
        self._add_result_response_content(result, response_content)
        self._add_extra_response_content(response_content)
        return response_content

    def _default_response_content(self):
        return {}

    def _add_result_response_content(self, result, response_content):
        data = self._trimmed_result(result)
        response_content.update(data)
        return response_content

    def _add_extra_response_content(self, response_content):
        data = self._extra_data()
        response_content.update(data)
        return response_content

    def _notify_post_request(self, request, interaction, result):
        """
        Override this method if you want this handler to notify its execution to an announcer
        """
        pass

    def _extra_data(self) -> Dict[object, object]:
        """
            Override this method if you want to add additional data to the response.
             examples: App notifications, session status, the API version, endpoint deprecation warnings.
        """
        return {}

    def _result_with_error(self, error):
        result = Result()
        result.add_error(error)
        return result



from unittest import TestCase

from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.application.interactions.validators.generic import NotNone
from appyx.layers.domain.exceptions.message import MessageMissingArgumentsException, \
    MessageWithInvalidArgumentsException
from appyx.layers.domain.messages.message import Message
from appyx.layers.domain.messages.type_of_message import TypeOfMessage


class MessageTest(TestCase):
    def test_it_can_be_displayed_as_text_when_it_does_not_have_arguments(self):
        # Given a message created from a type of message
        operation_failed_text = 'The operation has failed for undisclosed reasons.'
        type_of_message = TypeOfMessage(meaning='operation_failed', text_template=operation_failed_text)
        message = Message(type_of_message)

        # When it is queried for its text presentation
        message_text = message.text()
        message_string = str(message)

        # Then it returns the type's text
        self.assertEqual(operation_failed_text, message_text)
        self.assertEqual(operation_failed_text, message_string)

    def test_it_can_be_displayed_as_text_when_its_parameters_have_arguments(self):
        # Given a message
        parametrized_text = ["There have been ", Parameter(name="amount", example_argument=4), " accidents this year."]
        type_of_message = TypeOfMessage.new_with_parametrized_text('yearly_workplace_accidents', parametrized_text)
        message = Message(type_of_message, arguments={"amount": 15})

        # When it is queried for its text presentation
        message_text = message.text()
        message_string = str(message)

        # Then it returns a human-readable text using the given argument values for its parameters
        self.assertEqual(message_text, "There have been 15 accidents this year.")
        self.assertEqual(message_string, "There have been 15 accidents this year.")

    def test_it_can_be_instantiated_with_argument_names_as_keywords(self):
        # Given a type of message with parameters
        avoidable_amount = Parameter(name="avoidable_amount", example_argument=4)
        total_amount = Parameter(name="total_amount", example_argument=5)

        parametrized_text = ["Approximately ", avoidable_amount, " out of ", total_amount, " accidents were avoidable."]
        type_of_message = TypeOfMessage.new_with_parametrized_text('yearly_workplace_accidents', parametrized_text)

        # When a message of that type is instantiated with arguments as keywords
        message = Message(type_of_message, avoidable_amount=5, total_amount=7)

        # Then it can use them in its text presentation
        self.assertEqual("Approximately 5 out of 7 accidents were avoidable.", message.text())

    def test_it_can_combine_dict_arguments_and_keyword_arguments(self):
        # Given a type of message with parameters
        avoidable_amount = Parameter(name="avoidable_amount", example_argument=4)
        total_amount = Parameter(name="total_amount", example_argument=5)

        parametrized_text = ["Approximately ", avoidable_amount, " out of ", total_amount, " accidents were avoidable."]
        type_of_message = TypeOfMessage.new_with_parametrized_text('yearly_workplace_accidents', parametrized_text)

        # When a message of that type is instantiated with arguments as keywords
        message = Message(type_of_message, arguments={"total_amount": 7}, avoidable_amount=5)

        # Then it can use them in its text presentation
        self.assertEqual("Approximately 5 out of 7 accidents were avoidable.", message.text())

    def test_it_ignores_unused_arguments(self):
        # Given a type of message with parameters
        avoidable_amount = Parameter(name="avoidable_amount", example_argument=4)
        total_amount = Parameter(name="total_amount", example_argument=5)

        parametrized_text = ["Approximately ", avoidable_amount, " out of ", total_amount, " accidents were avoidable."]
        type_of_message = TypeOfMessage.new_with_parametrized_text('yearly_workplace_accidents', parametrized_text)

        # When a message of that type is instantiated with arguments as keywords
        message = Message(type_of_message, avoidable_amount=5, total_amount=7, roberto="roberto")

        # Then it can use them in its text presentation
        self.assertEqual("Approximately 5 out of 7 accidents were avoidable.", message.text())

    def test_it_cannot_be_instantiated_without_required_arguments(self):
        # Given a type of message
        parametrized_text = ["There have been ", Parameter(name="amount", example_argument=4), " accidents this year."]
        type_of_message = TypeOfMessage.new_with_parametrized_text('yearly_workplace_accidents', parametrized_text)
        raised_exception = None

        # When it is attempted to create a message with that type and with missing arguments
        try:
            Message(type_of_message)
        except MessageMissingArgumentsException as exception:
            raised_exception = exception

        # Then it fails to create and raises an exception informing the missing arguments
        self.assertEqual("El parámetro amount es obligatorio", str(raised_exception))

    def test_it_cannot_be_instantiated_with_invalid_arguments(self):
        # Given a type of message
        parametrized_text = ["There have been ", Parameter(name="amount", validators=[NotNone()], example_argument=4),
                             " accidents this year."]
        type_of_message = TypeOfMessage.new_with_parametrized_text('yearly_workplace_accidents', parametrized_text)
        raised_exception = None

        # When it is attempted to create a message with that type and with invalid arguments
        try:
            Message(type_of_message, arguments={"amount": None})
        except MessageWithInvalidArgumentsException as exception:
            raised_exception = exception

        # Then it fails to create and raises an exception informing the missing arguments
        self.assertEqual(
            "Parameter \"amount\" is not valid with argument \"None\". Reasons: Este campo no puede estar vacío",
            str(raised_exception)
        )



class MessageMissingArgumentsException(Exception):
    """
    This object models a lack of required arguments in a message.
    """
    pass


class MessageWithInvalidArgumentsException(Exception):
    """
    This object models a use of invalid arguments in a message.
    """
    pass



class HttpParameterLocation:
    def get_argument_from_request(self, parameter, request, kwargs, default=None):
        raise NotImplementedError('Subclass responsibility')

    @classmethod
    def path(cls):
        return PathHttpParameter()

    @classmethod
    def data(cls):
        return DataHttpParameter()

    @classmethod
    def query(cls):
        return QueryHttpParameter()

    @classmethod
    def file(cls):
        return FileHttpParameter()

    @classmethod
    def auth(cls):
        return AuthHttpParameter()


# TODO: add support to path parameters and implement this. This needs that the handler is decoupled from the APIView
class PathHttpParameter(HttpParameterLocation):
    def get_argument_from_request(self, parameter, request, kwargs, default=None):
        return kwargs.get(parameter.name(), default)


class DataHttpParameter(HttpParameterLocation):
    def get_argument_from_request(self, parameter, request, kwargs, default=None):
        return request.data.get(parameter.name(), default)


class QueryHttpParameter(HttpParameterLocation):
    def get_argument_from_request(self, parameter, request, kwargs, default=None):
        return request.GET.get(parameter.name(), default)


class FileHttpParameter(HttpParameterLocation):
    def get_argument_from_request(self, parameter, request, kwargs, default=None):
        return request.FILES.get(parameter.name(), default)


class AuthHttpParameter(HttpParameterLocation):
    def get_argument_from_request(self, parameter, request, kwargs, default=None):
        return request.auth



from appyx.layers.domain.validators.object_validator import ObjectValidator


class DictValidator(ObjectValidator):
    def _validate_with_result(self, an_object, result):
        raise NotImplementedError()

    def validate_has_exact_keys(self, dictionary, expected_keys, result):
        return self._assert_keys_defined(dictionary, expected_keys, result)

    def _assert_keys_defined(self, dictionary, expected_keys, result):
        for key in expected_keys:
            self._assert_key_defined(dictionary, key, result)
        expected_keys_size = len(expected_keys)
        dictionary_keys_size = len(dictionary.keys())
        if dictionary_keys_size != expected_keys_size:
            result.add_error(f"Dictionary expected to have {expected_keys_size} keys has {dictionary_keys_size} keys")
        return result

    def _assert_key_defined(self, dictionary, expected_key, result):
        if expected_key not in dictionary:
            result.add_error(f"Key {expected_key} missing from dictionary")
        return result

    def validate_key_defined(self, dictionary, expected_key, result, validators=None):
        validators = validators or []
        result = self._assert_key_defined(dictionary, expected_key, result)
        if result.has_errors():
            return result

        value = dictionary[expected_key]
        for validator in validators:
            validator.validate(value, result)

        return result

    def validate_keys_defined(self, dictionary, expected_keys, result, validators=None):
        return self.validate_has_at_least_keys(dictionary, expected_keys, result, validators)

    def validate_has_at_least_keys(self, dictionary, expected_keys, result, validators=None):
        validators = validators or []
        for expected_key in expected_keys:
            self.validate_key_defined(dictionary, expected_key, result, validators)

        return result

    def validate_values_with_result(self, dictionary, validators, result):
        values = dictionary.values()
        for value in values:
            for validator in validators:
                validator.validate_with_result(value, result)

        return result

    def validate_keys_with_result(self, dictionary, validators, result):
        keys = dictionary.keys()
        for key in keys:
            for validator in validators:
                validator.validate_with_result(key, result)

        return result

    def validate_keys_included_in(self, dictionary, valid_keys, result=None):
        result = self._defined_result(result)
        for key in dictionary:
            self.validate_key_in(key, valid_keys, result)
        return result

    def validate_key_in(self, key, valid_keys, result):
        return self._assert_key_in(key, valid_keys, result)

    def _assert_key_in(self, key, expected_keys, result):
        if key not in expected_keys:
            result.add_error(f"Key {key} is not included in {expected_keys}")
        return result



from appyx.layers.domain.errors import BaseErrorFactory


class Result:
    """This class models the result from the validate or execute of an Interaction or Operation

    It has an object, which is set by the interaction/operation. It is the product of the execution.
    Example:
        1) Adding a new user would result in the result of that operation to have the user in the result object
        2) Closing a poll may return an object containing the boolean indicating success or failure of the poll,
        and a list of positive voters and negative voters. All of these in a dictionary.
    It has errors, which contain a list of problems in a human readable way. It is empty if the execution was successful

    Note that objects that return a Result should not throw/raise exceptions.
    Because any errors should be contained and explained in the Result.
    """

    def __init__(self):
        self._errors = []
        # TODO: Add warnings
        self._object = None

    def get_object(self):
        return self._object

    def set_object(self, object_to_set):
        self._object = object_to_set

    def is_successful(self):
        return len(self._errors) == 0

    def errors(self):
        return self._errors

    def error_codes(self):
        return [error.code() for error in self._errors]

    def has_error_with_code(self, error_code):
        return error_code in self.error_codes()

    def has_at_least_one_error_with_codes(self, error_codes):
        return any([error_code for error_code in error_codes if error_code in self.error_codes()])

    def add_error(self, an_object):
        if type(an_object) is list:
            self.add_errors(an_object)
            return

        if type(an_object) is str:
            error = BaseErrorFactory().simple_error(an_object)
        elif isinstance(an_object, Exception):
            error = BaseErrorFactory().unexpected_error(an_object)
        else:
            error = an_object

        self._errors.append(error)

    def add_errors(self, errors):
        for error in errors:
            self.add_error(error)

    def add_errors_from(self, a_result):
        self.add_errors(a_result.errors())

    def has_errors(self):
        return not self.is_successful()

    def replace_error(self, old_error_code, new_error):
        self._errors = [new_error if error.code() == old_error_code else error for error in self._errors]

    # Convenience method
    def errors_as_string(self):
        return ', '.join([str(error) for error in self.errors()])

    def __str__(self):
        return f'Result ({self._object}) with errors {self._errors}'

    def add_warning(self, text):
        # TODO: warning support in Result
        pass

    def delete_repeated_errors(self):
        errors_set = set(self.errors())
        self._errors = list(errors_set)

    @classmethod
    def with_error(cls, error):
        result = cls()
        result.add_error(error)
        return result



from appyx.layers.domain.messages.exceptions import ParameterNotFoundException


class HttpParameter:  # inherit from parameter? we also need validators? -> yes
    def __init__(self, name, location, example):
        self._name = name
        self._location = location
        self._example = example

    def name(self):
        return self._name

    def location(self):
        return self._location

    def example(self):
        return self._example

    def get_argument_from_request(self, request, kwargs, default=None):
        return self._location.get_argument_from_request(self, request, kwargs, default)


class HandlerParameters:  # Check how this object's protocol relates to SetOfParameters.
    def __init__(self, initial_parameters=None, initial_required_parameter_names=None):
        self._parameters = initial_parameters or []
        self._required_parameter_names = initial_required_parameter_names or set()

    def required_parameter_names(self):
        return self._required_parameter_names

    def parameters(self):
        return self._parameters

    def get_argument_named(self, request, kwargs, parameter_name, default=None):
        parameter = self._parameter_named(parameter_name)
        return parameter.get_argument_from_request(request, kwargs, default)

    def _parameter_named(self, parameter_name):
        for parameter in self.parameters():
            if parameter_name == parameter.name():
                return parameter
        raise ParameterNotFoundException(parameter_name)



class EarlyReturnBecauseOfErrorsException(Exception):
    """
    This object models an early return from the operation logic because of errors in its execution.
    It is used to stop execution of the logic from inside an object and return no matter where we are.

    Since a method from an object can't delegate the return from itself to a method it invokes
    throwing this exception and returning is used to mimic that behavior.
    """
    pass



class CommandLineParameter:
    def __init__(self, keyword, default_value):
        self._keyword = keyword
        self._default_value = default_value

    def value(self):
        import sys
        from getopt import getopt

        opts, args = getopt(sys.argv, "", [self._keyword + "="])
        for (opt, value) in opts:
            if opt == "--" + self._keyword:
                return value

        return self._default_value


class SettingsFileParameter:
    def __init__(self, settings, parameter_key):
        self._settings = settings
        self._parameter_key = parameter_key

    def value(self):
        return getattr(self._settings, self._parameter_key)


class EnvironmentVariableParameter:
    def __init__(self, var, default):
        self._var = var
        self._default = default

    def value(self):
        try:
            import os
            return os.environ[self._var]
        except KeyError:
            import logging
            logging.info("Using default value '{}' for environment variable {}".format(self._default, self._var))
            return self._default



from unittest import TestCase


class ApplicationTest(TestCase):

    def test_01_answers_the_same_business(self):
        from appyx.layers.application.test_environment import TestEnvironment
        first_asked_business = TestEnvironment().current_app().current_business()

        second_asked_business = TestEnvironment().current_app().current_business()

        self.assertIs(first_asked_business, second_asked_business)



from appyx.layers.application.exception_handler import AddExceptionToResult
from appyx.layers.domain.business import Business


class Application:
    """
    Models an application.
    Given a running environment it creates the resources that will be used by the business.
    It also defines and set up the interfaces with which external actors will interact with the business.

    An application also defines how to handle the exceptions those might rise for the resources the business uses.
    """

    def __init__(self, running_context=None, exception_handler=None) -> None:
        super().__init__()
        self._business = None
        self._general_exception_handler = exception_handler or self._default_exception_handler()
        if running_context is None:
            running_context = self.default_running_context()
        self._running_context = running_context

    def general_exception_handler(self):
        return self._general_exception_handler

    def set_general_exception_handler(self, exception_handler):
        self._general_exception_handler = exception_handler

    def _default_exception_handler(self):
        return AddExceptionToResult()

    def current_business(self):
        if self._business is None:
            self._business = self._new_business()
        return self._business

    def _new_business(self):
        return Business()

    def default_running_context(self):
        raise NotImplementedError('Subclass responsibility')



from appyx.layers.application.interactions.parameters.parameter import Parameter


class ErrorMessages:
    def argument_not_found_message(self, missing_argument_name):
        parametrized_text = [
            "Argument named: ",
            Parameter(name="argument_name", example_argument='password'),
            " not found"]
        return self._1_argument_message(parametrized_text, 'argument_not_found', missing_argument_name)

    def parameter_name_not_found(self, missing_parameter_name):
        parametrized_text = [
            "Parameter named: ",
            Parameter(name="parameter_name", example_argument='amount'),
            " not found"]
        return self._1_argument_message(parametrized_text, 'parameter_name_not_found', missing_parameter_name)

    def duplicated_parameter_name(self, duplicated_parameter_name):
        parametrized_text = [
            "Parameter ",
            Parameter(name="parameter_name", example_argument='amount'),
            " is already present"]
        return self._1_argument_message(parametrized_text, 'duplicated_parameter_name', duplicated_parameter_name)

    def parameter_example_missing(self, parameter):
        parametrized_text = [
            "Parameter with name ",
            Parameter(name="parameter_name", example_argument='password'),
            " missing an example argument"]
        return self._1_argument_message(parametrized_text, 'parameter_example_missing', parameter.name())

    def creation_error_message(self, class_been_instantiated, error_while_instantiating):
        parametrized_text = [
            "Cannot create instance of ",
            Parameter(name="class_name_been_instantiated", example_argument=self._example_class().__name__),
            ": ",
            Parameter(name="exception_or_message", example_argument=self._exception_message_example())]
        return self._2_argument_message(parametrized_text, 'argument_not_found',
                                        class_been_instantiated.__name__, error_while_instantiating)

    def _example_class(self):
        class Car:
            pass

        return Car

    def _exception_message_example(self):
        message = self._1_argument_message(["¡Felices ", Parameter('edad', example_argument=12), " años Maxi!"],
                                           'saludo_de_cumple', 16)
        from appyx.layers.domain.messages.exceptions import ExceptionMessage
        return ExceptionMessage(message)

    def _1_argument_message(self, parametrized_text, meaning, single_argument):
        """We assume a parametrized text with the form: [String, Parameter, ...]"""
        arguments = {parametrized_text[1].name(): single_argument}
        message = self._new_message(arguments, meaning, parametrized_text)
        return message

    def _2_argument_message(self, parametrized_text, meaning, argument1, argument2):
        """We assume a parametrized text with the form: [String, Parameter, String, Parameter, ...]"""
        arguments = {parametrized_text[1].name(): argument1, parametrized_text[3].name(): argument2}
        message = self._new_message(arguments, meaning, parametrized_text)
        return message

    def _new_message(self, arguments, meaning, parametrized_text):
        from appyx.layers.domain.messages.message import Message
        from appyx.layers.domain.messages.type_of_message import TypeOfMessage
        type_of_message = TypeOfMessage.new_with_parametrized_text(
            meaning=meaning,
            parametrized_text=parametrized_text)
        message = Message(type_of_message=type_of_message,
                          arguments=arguments)
        return message



from __future__ import annotations
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.domain.errors import BaseErrorFactory
from appyx.layers.domain.result import Result


class ParametersValidator:
    """
        I validate the command's parameters
    """

    def validate(self, parameters_and_arguments, result) -> Result:
        validation_result = self._validation_result(parameters_and_arguments)
        if validation_result.is_successful():
            self._on_valid_argument(parameters_and_arguments, validation_result, result)
        else:
            self._on_invalid_argument(parameters_and_arguments, validation_result, result)
        return result

    def _validation_result(self, parameters_and_arguments) -> Result:
        raise NotImplementedError("subclass responsibility")

    def _on_valid_argument(self, parameters_and_arguments, validation_result, result):
        pass

    def _on_invalid_argument(self, parameters_and_arguments, validation_result, result):
        result.add_error(BaseErrorFactory().parameters_validation_error(parameters_and_arguments,
                                                                        validation_result.errors()))


class NoneValidation(ParametersValidator):
    def validate(self, parameters_and_arguments, result):
        return result


class RequiredParameters(ParametersValidator):
    """
        I validate the required command's parameters
    """

    def __init__(self, required_parameters: List[Parameter]):
        super().__init__()
        self._required_parameters = required_parameters

    def _validation_result(self, parameters_and_arguments):
        validation_result = Result()
        for parameter_name in self._required_parameters:
            if parameter_name not in parameters_and_arguments.keys():
                validation_result.add_error(u'The parameter %s is required' % parameter_name)
        return validation_result

    @classmethod
    def new_with(cls, required_parameters):
        return cls(required_parameters)


class AtLeastOneParameterRequired(ParametersValidator):
    """
        I validate that the command has at least one of its required parameters.
    """

    def __init__(self, required_parameters):
        super().__init__()
        self._required_parameters = required_parameters

    def _validation_result(self, parameters_and_arguments):
        validation_result = Result()
        for parameter_name in self._required_parameters:
            if parameter_name in parameters_and_arguments.keys():
                return validation_result
        validation_result.add_error(
            u'At least one of the following parameters is required: {0}.'.format(u', '.join(self._required_parameters)))
        return validation_result


class AtLeastOneParameterNotNoneRequired(ParametersValidator):
    """
        I validate that the command has at least one of its required parameters set as not None.
    """

    def __init__(self, required_parameters):
        super().__init__()
        self._required_parameters = required_parameters

    def _validation_result(self, parameters_and_arguments):
        validation_result = Result()
        for parameter_name in self._required_parameters:
            if parameter_name in parameters_and_arguments.keys():
                argument = parameters_and_arguments[parameter_name]
                if argument is not None:
                    return validation_result
        validation_result.add_error(
            u'At least one of the following parameters is required and not none: {0}.'.format(
                u', '.join(self._required_parameters)))
        return validation_result


class ParameterValidator:
    """
    Models a validator for a parameter.
    """

    def validate(self, argument: object, result: Result = None, parameter_name: str = None) -> bool:
        if argument is None:
            return True

        result = self.validate_with_result(argument, result, parameter_name)
        return result.is_successful()

    def validate_with_result(self, argument: object, result: Result = None, parameter_name: str = None) -> Result:
        if result is None:
            result = Result()

        return self._validate_with_result(argument, result, parameter_name)

    def _validate_with_result(self, argument, result, parameter_name):
        validation_result = self._validation_result(argument)
        if validation_result.is_successful():
            self._on_valid_argument(argument, validation_result, result, parameter_name)
        else:
            self._on_invalid_argument(argument, validation_result, result, parameter_name)
        return result

    def _on_invalid_argument(self, argument, validation_result, result, parameter_name):
        if parameter_name is None:
            parameter_name = self.__class__.__name__

        result.add_error(BaseErrorFactory().parameter_validation_error(parameter_name, argument,
                                                                       validation_result.errors()))

    def _validation_result(self, argument: object) -> Result:
        raise NotImplementedError('subclass responsibility')

    def _on_valid_argument(self, argument: object, validation_result: Result, result: Result, parameter_name: str):
        pass



from appyx.layers.application.application import Application
from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.application.interactions.parameters.set_of_parameters import SetOfParameters
from appyx.layers.application.interactions.validators.generic import IsOneOfThese, IsBoolean


class PollsApplication(Application):
    CLOCK_FAKE = 'FAKE'
    CLOCK_SYSTEM = 'SYSTEM'
    QUESTIONS_EPHEMERAL = 'EPHEMERAL'  # or TRANSIENT
    QUESTIONS_PERSISTENT = 'PERSISTENT'

    @classmethod
    def running_context_test_dev(cls):
        initial_parameters = [
            Parameter("clock_behavior", [IsOneOfThese(['FAKE, SYSTEM'])]),
            Parameter("questions_volatility", [IsOneOfThese(['EPHEMERAL', 'PERSISTENT'])]),
            Parameter("run_in_parallel", [IsBoolean()]),
        ]

        default_arguments = {"clock_behavior": 'FAKE'}
        context = SetOfParameters(None,
                                  initial_parameters=initial_parameters,
                                  initial_required_parameter_names=["questions_volatility", "run_in_parallel"],
                                  default_arguments=default_arguments)
        return context

    @classmethod
    def running_context_production(cls):
        initial_parameters = [
            Parameter("clock_behavior", [IsOneOfThese(['FAKE, SYSTEM'])]),
            Parameter("questions_volatility", [IsOneOfThese(['EPHEMERAL', 'PERSISTENT'])]),
        ]

        default_arguments = {"clock_behavior": 'SYSTEM'}
        context = SetOfParameters(None,
                                  initial_parameters=initial_parameters,
                                  initial_required_parameter_names=["questions_volatility"],
                                  default_arguments=default_arguments)

        return context

    def _new_business(self):
        from polls.domain.poll_station import PollStation
        clock = self.clock(self._running_context.get_argument_named("clock_behavior"))
        questions = self._questions(self._running_context.get_argument_named("questions_repository"))
        business = PollStation(clock, questions)
        return business

    def default_running_context(self):
        return self.running_context_test_dev()

    def clock(self, clock_behavior):
        if clock_behavior == self.CLOCK_FAKE:
            from polls.domain.poll_station import FakeClock
            return FakeClock()
        if clock_behavior == self.CLOCK_SYSTEM:
            from polls.domain.poll_station import SystemClock
            return SystemClock()
        raise ValueError(f"Unsupported valid clock behavior: {clock_behavior}")

    def _questions(self, questions_volatility):
        if questions_volatility == self.QUESTIONS_EPHEMERAL:
            from polls.domain.poll_station import QuestionsArchive
            return QuestionsArchive()
        raise ValueError(f"Unsupported valid questions volatility: {questions_volatility}")



class SimpleErrorRenderer:
    """
    This class models the flattening of simple errors.
    I converts an error that follows the SimpleError protocol to a shallow object
    made up of basic, plain, shallow types/objects.
    """
    def render(self, error):
        result_dict = {
                "code": error.code(),
                "text": error.text()
            }
        return result_dict



from appyx.layers.application.environment import Environment, OneToOne
from polls import settings_stub
from polls.polls_application import PollsApplication
from appyx.layers.application.external_parameters import EnvironmentVariableParameter, CommandLineParameter, \
    SettingsFileParameter


class PollsEnvironment(Environment):
    TEST = 'TEST'
    PROD = 'PROD'

    def _application_class(self):
        return PollsApplication

    def _external_parameters_mapped_to_context_parameters(self):
        environment_parameter = EnvironmentVariableParameter('ENVIRONMENT', self.TEST)
        run_in_parallel_parameter = CommandLineParameter("run-in-parallel", default_value=False)
        questions_repository_parameter = SettingsFileParameter(settings_stub, "QUESTIONS_REPO")

        mappings = [OneToOne('environment', environment_parameter),
                    OneToOne('questions_repository', questions_repository_parameter)]
        if environment_parameter.value() == self.TEST:
            mappings.append(OneToOne('run_in_parallel', run_in_parallel_parameter))
        return mappings

    def _new_application_context(self):
        environment_parameter = EnvironmentVariableParameter('ENVIRONMENT', self.TEST)
        if environment_parameter.value() == self.PROD:
            return PollsApplication.running_context_production()
        else:
            return PollsApplication.running_context_test_dev()



from appyx.layers.domain.validators.object_validator import ObjectValidator


class Presenter:
    """
        I model the cropping of information made to a domain object when it is exposed to a client.
    """
    @classmethod
    def format_date(cls, a_date, a_format=None):
        if a_format is None:
            return cls.format_date(a_date, a_format=cls.default_date_format())
        return a_date.strftime(a_format)

    @classmethod
    def format_datetime(cls, a_datetime, a_format=None):
        if a_format is None:
            return cls.format_datetime(a_datetime, a_format=cls.default_datetime_format())
        return a_datetime.strftime(a_format)

    @classmethod
    def default_date_format(cls):
        return "%Y/%m/%d"

    @classmethod
    def default_datetime_format(cls):
        return ObjectValidator.default_datetime_format()



import re


class AppyxEndpoint:
    def __init__(self, handler, route, name, interface) -> None:
        super().__init__()
        self._route = route
        self._handler = handler
        self._name = name
        self._interface = interface

    def interface(self):
        return self._interface

    def get_authentication_methods(self):
        return self.interface().authentication_methods_for_endpoint_named(self.name())

    def route(self):
        return self._route

    def route_with_variable_replacements(self, kwargs=None):
        if kwargs is None:
            kwargs = {}

        def replace(m):
            variable_name = m.group(1)
            return str(kwargs[variable_name])

        pattern = self._pattern_for_variables_in_url()
        replaced_route = re.sub(pattern, replace, self._route)
        return replaced_route

    def handler(self):
        return self._handler

    def http_method_to_handler(self):
        http_method_name = self.http_method().lower()
        return {http_method_name: self.handler()}

    def name(self):
        return self._name

    def http_method(self):
        # This is coupled with Django :(
        return self.handler().allowed_methods[0]

    def _pattern_for_variables_in_url(self):
        return r"<[a-z]+:(?P<variable_name>[\_\-a-zA-Z0-9]+)>"



class Mapping:
    def cast_on(self, set_of_parameters):
        raise NotImplementedError('Subclass responsibility')


class OneToOne(Mapping):
    def __init__(self, parameter_name, parameter_source):
        self._parameter_name = parameter_name
        self._parameter_source = parameter_source

    def cast_on(self, set_of_parameters):
        set_of_parameters.set_argument_value(self._parameter_name, self._parameter_source.value())


class Environment:
    CURRENT_APP = None

    def current_app(self):
        if self.__class__.CURRENT_APP is None:
            self.__class__.CURRENT_APP = self._new_application()
        return self.__class__.CURRENT_APP

    def _new_application(self):
        context = self._running_context()
        app_class = self._application_class()
        return app_class(context)

    def reset_current_app(self):
        self.__class__.CURRENT_APP = None

    def _application_class(self):
        raise NotImplementedError('Subclass responsibility')

    def _running_context(self):
        context = self._new_application_context()
        mappings = self._external_parameters_mapped_to_context_parameters()
        self._validate_mappings_cover_all_context_variables(context, mappings)
        self._validate_that_all_context_parameters_have_a_defined_mapping(context, mappings)
        for mapping in mappings:
            mapping.cast_on(context)
        context.validate()
        return context

    def _external_parameters_mapped_to_context_parameters(self):
        raise NotImplementedError('Subclass responsibility')

    def _new_application_context(self):
        raise NotImplementedError('Subclass responsibility')

    def _validate_mappings_cover_all_context_variables(self, context, mappings):
        """not implemented yet"""
        pass

    def _validate_that_all_context_parameters_have_a_defined_mapping(self, context, mappings):
        """not implemented yet"""
        pass



import unittest

from appyx import Result
from appyx.layers.application.interactions.validators.generic import IsAPositiveInteger, IsString
from appyx.layers.interface.validators.dict_validator import DictValidator


class DictValidatorTest(unittest.TestCase):
    def test_empty_dict_has_keys_included(self):
        dictionary = {}
        validator = DictValidator()

        result = validator.validate_keys_included_in(dictionary, [])

        self.assertTrue(result.is_successful())

    def test_dict_has_keys_included(self):
        dictionary = {'key1': 'value1'}
        validator = DictValidator()

        result = validator.validate_keys_included_in(dictionary, ['key1', 'key2'])

        self.assertTrue(result.is_successful())

    def test_when_dict_has_keys_not_included_an_error_is_answered(self):
        dictionary = {'key3': 'value1'}
        validator = DictValidator()

        result = validator.validate_keys_included_in(dictionary, ['key1', 'key2'])

        self.assertTrue(result.has_errors())
        self.assertTrue(result.has_error_with_code('simple_error_code'))
        self.assertEqual(result.errors()[0].text(), 'Key key3 is not included in [\'key1\', \'key2\']')

    def test_we_can_validate_the_dictionary_keys(self):
        dictionary = {1: 'un valor', 2: 3, 'key invalida': 'una string'}
        validator = DictValidator()

        result = validator.validate_keys_with_result(dictionary, [IsAPositiveInteger()], Result())

        self.assertTrue(result.has_errors())
        self.assertEqual(1, len(result.errors()))

    def test_we_can_validate_the_dictionary_values(self):
        dictionary = {1: 'un valor', 2: 3, 'key invalida': 'una string'}
        validator = DictValidator()

        result = validator.validate_values_with_result(dictionary, [IsString()], Result())

        self.assertTrue(result.has_errors())
        self.assertEqual(1, len(result.errors()))

    def test_we_can_validate_the_dictionary_has_at_least_some_keys(self):
        dictionary = {1: 'un valor', 2: 3, 'key textual': 'una string'}
        validator = DictValidator()

        result = validator.validate_keys_defined(dictionary, [1, 'key textual'], Result())

        self.assertTrue(result.is_successful())

    def test_we_can_validate_the_dictionary_doesnt_have_at_least_some_keys(self):
        dictionary = {1: 'un valor', 2: 3, 'key textual': 'una string'}
        validator = DictValidator()

        result = validator.validate_keys_defined(dictionary, [1, 2, 3], Result())

        self.assertTrue(result.has_errors())

    def test_we_can_validate_the_dictionary_has_exactly_some_keys(self):
        dictionary = {1: 'un valor', 2: 3, 'key textual': 'una string'}
        validator = DictValidator()

        result = validator.validate_has_exact_keys(dictionary, [1, 2, 'key textual'], Result())

        self.assertTrue(result.is_successful())

    def test_we_can_validate_the_dictionary_doesnt_have_exactly_some_keys(self):
        dictionary = {1: 'un valor', 2: 3, 'key textual': 'una string'}
        validator = DictValidator()

        result = validator.validate_has_exact_keys(dictionary, [1, 2], Result())

        self.assertTrue(result.has_errors())

    def test_we_can_validate_the_dictionary_has_a_key_and_its_value_is_valid(self):
        dictionary = {1: 'un valor', 2: 3, 'key textual': 'una string'}
        validator = DictValidator()

        result = validator.validate_key_defined(dictionary, 2, Result(), [IsAPositiveInteger()])

        self.assertTrue(result.is_successful())

    def test_sadly_a_dict_validator_cannot_validate_a_dictionary(self):
        dictionary = {1: 'un valor', 2: 3, 'key textual': 'una string'}
        validator = DictValidator()
        not_implemented_error_raised = False

        try:
            result = validator.validate(dictionary)
        except NotImplementedError:
            not_implemented_error_raised = True

        self.assertTrue(not_implemented_error_raised)



from unittest import TestCase

from appyx.layers.application.environment import Environment


class TestEnvironment(Environment):

    def _application_class(self):
        from appyx.layers.application.interactions.parameters.test_command_parameters import TestApp
        return TestApp

    def _running_context(self):
        return 'a running context'


class EnvironmentTest(TestCase):

    def test_01_answers_the_same_application(self):
        first_asked_app = TestEnvironment().current_app()

        second_asked_app = TestEnvironment().current_app()

        self.assertIs(first_asked_app, second_asked_app)



from typing import List

from appyx.layers.application.interactions.validators.parameters import ParameterValidator
from appyx.layers.domain.result import Result


class Parameter:
    """
    Models one requirement of an interaction or operation.
    It has a name and a way to tell if it is valid (i.e. the requirement fulfilled) by an argument.

    Example:
        If I have a money deposit operation, with two parameters: amount and currency
        the parameter amount would have as requirement to be a number, maybe non-negative or positive
        the parameter currency would have as requirement to be one of the possible currencies in the system.
        The arguments 100 and "USD" would satisfy those parameters, they would be valid.
        The arguments "one-hundred" and "MoneyMoneyMoney" would not.
    """

    def __init__(self, name: str, validators: List[ParameterValidator] = None, description='', example_argument=None):
        super().__init__()
        self._name = name
        self._validators = validators if type(validators) is list else []
        self._description = description
        self._example_argument = example_argument

    def example_argument(self):
        return self._example_argument

    def validate(self, argument, result=None) -> Result:
        if result is None:
            result = Result()

        validators_result = Result()
        for validator in self._validators:
            validator.validate(argument, validators_result, self._name)

        result.add_errors(validators_result.errors())
        return result

    def name(self):
        return self._name

    def description(self):
        return self._description

    def has_example_argument(self):
        return self._example_argument is not None



class Message:
    def __init__(self, type_of_message, arguments=None, **kwargs) -> None:
        from appyx.layers.domain.errors import BaseErrorFactory
        self._error_factory = BaseErrorFactory()
        self._type = type_of_message

        if arguments is None:
            arguments = {}
        arguments.update(kwargs)
        self._arguments = arguments

        self._assert_arguments_are_present_and_valid()

    def __str__(self):
        return self.text()
    
    def text(self):
        return self._type.text_template().format(**self._arguments)

    # --- private methods ---

    def _assert_arguments_are_present_and_valid(self):
        from appyx.layers.domain.exceptions.message import MessageMissingArgumentsException, \
            MessageWithInvalidArgumentsException
        result = self._type.validate_arguments(self._arguments)

        for error in result.errors():
            error_text = error.text()
            if error.code() == self._error_factory.parameter_missing_error_code():
                raise MessageMissingArgumentsException(error_text)
            if error.code() == self._error_factory.parameter_validation_error_code():
                raise MessageWithInvalidArgumentsException(error_text)
